/* (C) 1999
 * ESI/Supinfo
 * 
 * libparam: fonctions pour le gestion de paramtres et de fichiers de config.
 * (C) HPicollo 1998
 * Compltement pomp sur gnome-config.c , par Miguel de Icaza
 * (C) 1993, 1994, 1997 The Free Software Foundation
 *
 * -------------------------------------------
 * cfg.c:
 * Gestion des fichiers de configuration
 * -------------------------------------------
 * Zimmermann Jrmie	(j_zimmer@supinfo.com)
 * Praszalowicz Stefan	(s_prasza@supinfo.com)
 * -------------------------------------------
 * 
 * This is Free Software (see COPYING for details)
 */

#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>


/* -=-=-=-=-=--=-=-=-=-=-= GLOBAL =-=-=-=-=--=-=-=-=-=-=- */


#define STRSIZE 200
#define overflow (next == &CharBuffer [STRSIZE-1])

enum {
   FirstBrace,
     OnSecHeader,
     IgnoreToEOL,
     KeyDef,
     KeyDefOnKey,
     KeyValue
};

typedef struct {
   gchar * pp_path;
   gchar * pp_file;
   gchar * pp_section;
   gchar * pp_key;
} PPath;

typedef enum {
   SET,
     GET
} accesstype;

typedef struct cfg_keys_s {
   gchar * key_name;
   gchar * value;
   struct cfg_keys_s * link;
} cfg_keys;

typedef struct cfg_section_s {
   gchar * section_name;
   cfg_keys * keys;
   struct cfg_section_s * link;
} cfg_section;

typedef struct cfg_profile_s {
   gchar * filename;
   cfg_section * section;
   struct cfg_profile_s * link;
} cfg_profile;

static cfg_profile * Base;
static cfg_profile * Current;
static gchar homedir[1024] = "\0";

/* -=-=-=-=-=--=-=-=-=-=-= PRIVATE =-=-=-=-=--=-=-=-=-=-=- */

/* Path:
 * /fichier/section/clef
 * => HOMEDIR/fichier.cfg
 * => section
 * => clef
 */
static PPath * cfg_parse_path(const gchar * str)
{
	PPath * pp = g_malloc(sizeof(PPath));
	gchar * end;

	g_assert(str != NULL);
	g_assert(pp != NULL);

	if (homedir[0] == '\0')
	{
		g_free(pp);
		g_warning( ("Le rpertoire des fichiers de config. n'a pas t choisi."));
		return (PPath*) NULL;
	}

	if (*str != '/') {
		g_free(pp);
		g_warning( ("Le `path' %s est mal form (cfg parse path)"), str);
		return (PPath*) NULL;
	}

	pp->pp_path = g_strdup(str);

	end = pp->pp_path + strlen(pp->pp_path);

	while(end > pp->pp_path) {
		end--;
		if (*end == '/') {
			*end = 0;
			pp->pp_key = end + 1;
			break;
		}
	}

	while (end > pp->pp_path) {
		end--;
		if (*end == '/') {
			*end = 0;
			pp->pp_section = end + 1;
			break;
		}
	}

	pp->pp_file = g_strconcat(homedir, pp->pp_path, "rc", NULL);

	return pp;
}

static void cfg_release_path(PPath *pp)
{
	g_free(pp->pp_path);
	g_free(pp->pp_file);
	g_free(pp);
}


static gint cfg_is_loaded(const gchar * filename, cfg_section ** section)
{
	cfg_profile * p = Base;

	while (p) {
		if (!g_strcasecmp(filename, p->filename)) {
			Current = p;
			*section = p->section;
			return TRUE;
		}
		p = p->link;
	}
	return FALSE;
}

static gchar * cfg_decode_string(gchar *s)
{
	gchar *p = g_malloc (strlen (s) + 1);
	gchar *q = p;

	do {
		if (*s == '\\'){
			switch (*(++s)){
			 case 'n':
				*p++ = '\n';
				break;
			 case '\\':
				*p++ = '\\';
				break;
			 case 'r':
				*p++ = '\r';
				break;
			 default:
				*p++ = '\\';
				*p++ = *s;
			}
		} else
		  *p++ = *s;
	} while (*s++);
	return q;
}

static gchar * cfg_encode_string(gchar *s)
{
	gchar *return_value, *p = s;
	int len = 0;

	while (*p){
		len++;
		if (*p == '\n' || *p == '\\' || *p == '\r' || *p == '\0')
		  len++;
		p++;
	}
	return_value = p = (gchar *) g_malloc (len + 1);
	if (!return_value)
	  return 0;
	do {
		switch (*s){
		 case '\n':
			*p++ = '\\';
			*p++ = 'n';
			break;
		 case '\r':
			*p++ = '\\';
			*p++ = 'r';
			break;
		 case '\\':
			*p++ = '\\';
			*p++ = '\\';
			break;
		 default:
			*p++ = *s;
		}
	} while (*s++);
	return return_value;
}

static cfg_section * cfg_load(const gchar * file)
{
	FILE *f;
	gint state;
	cfg_section *section = 0;
	gchar CharBuffer [STRSIZE];
	gchar *next = "";
	gint c;

	if ((f = fopen (file, "r"))==NULL)
	  return NULL;

	state = FirstBrace;
	while ((c = getc (f)) != EOF){
		if (c == '\r')		/* Ignore Carriage Return */
		  continue;

		switch (state){

		 case OnSecHeader:
			if (c == ']' || overflow){
				*next = '\0';
				next = CharBuffer;
				section->section_name = g_strdup (CharBuffer);
				g_strstrip(section->section_name);
				state = IgnoreToEOL;
			} else
			  *next++ = c;
			break;

		 case IgnoreToEOL:
			if (c == '\n'){
				state = KeyDef;
				next = CharBuffer;
			}
			break;

		 case FirstBrace:
		 case KeyDef:
		 case KeyDefOnKey:
			if (c == '['){
				cfg_section *temp;

				temp = section;
				section = (cfg_section *) g_malloc (sizeof (cfg_section));
				section->link = temp;
				section->keys = 0;
				state = OnSecHeader;
				next = CharBuffer;
				break;
			}
			/* On first pass, don't allow dangling keys */
			if (state == FirstBrace)
			  break;

			if ((c == ' ' && state != KeyDefOnKey) || c == '\t')
			  break;

			if (c == '\n' || overflow) /* Abort Definition */
			  next = CharBuffer;

			if (c == '=' || overflow){
				cfg_keys *temp;

				temp = section->keys;
				*next = '\0';
				section->keys = (cfg_keys *) g_malloc (sizeof (cfg_keys));
				section->keys->link = temp;
				section->keys->key_name = g_strdup (CharBuffer);
				g_strstrip(section->keys->key_name);
				state = KeyValue;
				next = CharBuffer;
			} else {
				*next++ = c;
				state = KeyDefOnKey;
			}
			break;

		 case KeyValue:
			if (overflow || c == '\n'){
				*next = '\0';
				section->keys->value = cfg_decode_string(CharBuffer);
				g_strstrip(section->keys->value);
				state = c == '\n' ? KeyDef : IgnoreToEOL;
				next = CharBuffer;

				/* printf ("[%s] (%s)=%s\n", section->section name,
					   * section->keys->key name, section->keys->value);
				 */
			} else
			  *next++ = c;
			break;

		} /* switch */

	} /* while ((c = getc (f)) != EOF) */

	if (c == EOF && state == KeyValue){
		*next = '\0';
		section->keys->value = cfg_decode_string(CharBuffer);
	}
	fclose (f);
	return section;
}

static void cfg_new_key (cfg_section *section, const gchar *key_name, const gchar *value)
{
	cfg_keys *key;

	key = (cfg_keys *) g_malloc (sizeof (cfg_keys));
	key->key_name = g_strdup (key_name);
	key->value   = g_strdup (value);
	key->link = section->keys;
	section->keys = key;
}

static const gchar * cfg_access(accesstype mode, const gchar * section_name,
				const gchar * key_name,
				const gchar * def, const gchar * filename)
{
	cfg_profile * New;
	cfg_section * section;
	cfg_keys * key;

	if (! cfg_is_loaded(filename, &section)) {
		New = g_malloc(sizeof(cfg_profile));
		New->link = Base;
		New->filename = g_strdup(filename);
		New->section = cfg_load(filename); // ?
		Base = New;
		section = New->section;
		Current = New;
	}

	/* Start search */
	for (; section; section = section->link) {
		if (section->section_name == NULL)
		  continue;
		if (g_strcasecmp (section->section_name, section_name))
		  continue;

		for (key = section->keys; key; key = key->link) {
			if (g_strcasecmp (key->key_name, key_name))
			  continue;
			if (mode == SET) {
				g_free (key->value);
				key->value = g_strdup (def);
			}
			return key->value;
		}

		/* No key found */
		if (mode == SET){
			cfg_new_key (section, key_name, def);
			return 0;
		}
	}

	/* Non existent section */
	if ((mode == SET) && def){
		section = (cfg_section *) g_malloc (sizeof(cfg_section));
		section->section_name = g_strdup (section_name);
		section->keys = 0;
		cfg_new_key (section, key_name, def);
		section->link = Current->section;
		Current->section = section;
	}
	return def;
}

static void cfg_dump_keys (FILE *profile, cfg_keys *p)
{
	if (!p)
	  return;
	cfg_dump_keys (profile, p->link);
	if (*p->key_name) {
		gchar *t = cfg_encode_string(p->value);
		fprintf (profile, "%s=%s\n", p->key_name, t);
		free (t);
	}
}

static void cfg_dump_sections (FILE *profile, cfg_section *p)
{
	if (!p)
	  return;
	cfg_dump_sections (profile, p->link);
	if (p->section_name [0]){
		fprintf (profile, "\n[%s]\n", p->section_name);
		cfg_dump_keys (profile, p->keys);
	}
}

static void cfg_dump_profile (cfg_profile *p)
{
	FILE *profile;

	if (!p)
	  return;
	cfg_dump_profile (p->link);

	if (p->filename[0] != (gchar) 0)
	  if ((profile = fopen (p->filename, "w")) != NULL){
		  cfg_dump_sections (profile, p->section);
		  fclose (profile);
	  }
}

static void cfg_free_keys (cfg_keys *p)
{
	if (!p)
	  return;
	cfg_free_keys (p->link);
	g_free (p->key_name);
	g_free (p->value);
	g_free (p);
}

static void cfg_free_sections (cfg_section *p)
{
	if (!p)
	  return;
	cfg_free_sections (p->link);
	cfg_free_keys (p->keys);
	g_free (p->section_name);
	p->link = 0;
	p->keys = 0;
	g_free (p);
}

static void cfg_free_profile (cfg_profile *p)
{
	if (!p)
	  return;
	cfg_free_profile (p->link);
	cfg_free_sections (p->section);
	g_free (p->filename);
	g_free (p);
}

/* -=-=-=-=-=--=-=-=-=-=-= PUBLIC =-=-=-=-=--=-=-=-=-=-=- */

gint cfg_set_home(gchar * newhome)
{
	struct stat buf;
	g_assert(newhome != NULL);

	if ((stat(newhome, &buf) != 0) || (!S_ISDIR(buf.st_mode)))
	{
		g_warning( ("Le rpertoire `%s' est invalide (cfg set home)"), newhome);
		return -1;
	}
	strncpy(homedir, newhome, 1024);
	return 0;
}

gchar * cfg_get_home(void)
{
	if (homedir[0] != '\0')
	  return homedir;
	return NULL;
}

void cfg_clean_file(const gchar *path)
{
	cfg_profile *p;
	PPath *pp;

	if (!path)
	  return;

	pp = cfg_parse_path (path);

	for (p = Base; p; p = p->link){
		if (strcmp (pp->pp_file, p->filename) != 0)
		  continue;
		cfg_free_sections (p->section);
		p->section = 0;
		p->filename [0] = 0;
		cfg_release_path(pp);
		return;
	}
	cfg_release_path (pp);
}

void cfg_sync (void)
{
	cfg_dump_profile (Base);
}

void cfg_drop_all(void)
{
	cfg_free_profile(Base);
}

/* /filename/section/key
 * /essai/main/essai: Fichier homedir/etc/essai.cfg Section Main, var Essai.
 */
gint cfg_get_int(gint defaut, gchar * fmt, ...)
{
	PPath * pp;
	const gchar *r;
	gchar defspace[100];
	gint v;
	va_list ap;
	gchar path[512];

	g_assert(fmt != NULL);

	va_start(ap, fmt);
	vsnprintf(path, 512, fmt, ap);
	va_end(ap);

	pp = cfg_parse_path(path);
	sprintf(defspace, "%d", defaut);
	r = cfg_access(GET, pp->pp_section, pp->pp_key, defspace, pp->pp_file);

	g_return_val_if_fail(r != NULL, 0);
	v = atoi(r);
	cfg_release_path(pp);
	return v;
}

gchar * cfg_get_str(gchar * defaut, gchar * fmt, ...)
{
	PPath * pp;
	const gchar *r;
	gchar * ret = NULL;
	va_list ap;
	gchar path[512];

	g_assert(fmt != NULL);

	va_start(ap, fmt);
	vsnprintf(path, 512, fmt, ap);
	va_end(ap);


	pp = cfg_parse_path(path);
	r = cfg_access(GET, pp->pp_section, pp->pp_key, defaut, pp->pp_file);

	if (r)
	  ret = g_strdup(r);
	cfg_release_path(pp);
	return ret;
}

int cfg_get_bool(gint defaut, gchar * fmt, ...)
{
	PPath * pp;
	const gchar *r;
	gchar defspace[100];
	gint v;

	va_list ap;
	gchar path[512];

	g_assert(fmt != NULL);

	va_start(ap, fmt);
	vsnprintf(path, 512, fmt, ap);
	va_end(ap);

	strcpy(defspace, defaut ? "true" : "false");

	pp = cfg_parse_path(path);
	r = cfg_access(GET, pp->pp_section, pp->pp_key, defspace, pp->pp_file);

	g_return_val_if_fail(r != NULL, FALSE);
	if (!g_strcasecmp(r, "true"))
	  v = TRUE;
	else
	  v = FALSE;

	cfg_release_path(pp);
	return v;
}

void cfg_set_str(const gchar *new_value, const gchar *fmt, ...)
{
	PPath *pp;
	const gchar *r;
	va_list ap;
	gchar path[512];

	g_assert(fmt != NULL);

	va_start(ap, fmt);
	vsnprintf(path, 512, fmt, ap);
	va_end(ap);

	pp = cfg_parse_path (path);
	r = cfg_access(SET, pp->pp_section, pp->pp_key, new_value, pp->pp_file);
	cfg_release_path (pp);
}

void cfg_set_int(gint new_value, const gchar *fmt, ...)
{
	PPath *pp;
	gchar intbuf [40];
	const gchar *r;
	va_list ap;
	gchar path[512];

	g_assert(fmt != NULL);

	va_start(ap, fmt);
	vsnprintf(path, 512, fmt, ap);
	va_end(ap);

	pp = cfg_parse_path (path);
	sprintf (intbuf, "%d", new_value);
	r = cfg_access(SET, pp->pp_section, pp->pp_key, intbuf, pp->pp_file);
	cfg_release_path(pp);
}

void cfg_set_bool(gint new_value, const gchar *fmt, ...)
{
	PPath *pp;
	const gchar *r;
	va_list ap;
	gchar path[512];

	g_assert(fmt != NULL);

	va_start(ap, fmt);
	vsnprintf(path, 512, fmt, ap);
	va_end(ap);

	pp = cfg_parse_path (path);
	r = cfg_access (SET, pp->pp_section, pp->pp_key,
			new_value ? "true" : "false", pp->pp_file);
	cfg_release_path (pp);
}

int cfg_section_exist(const gchar * fmt, ...)
{
	cfg_section *section;
	PPath *pp;
	cfg_profile * New;
	va_list ap;
	gchar path[512];

	g_assert(fmt != NULL);
	
	va_start(ap, fmt);
	vsnprintf(path, 512, fmt, ap);
	va_end(ap);

	pp = cfg_parse_path (path);

	if (!cfg_is_loaded (pp->pp_file, &section)) {
		New = g_malloc(sizeof(cfg_profile));
		New->link = Base;
		New->filename = g_strdup(pp->pp_file);
		New->section = cfg_load(pp->pp_file); // ?
		Base = New;
		section = New->section;
		Current = New;
	}

	for (; section; section = section->link) {
		if (g_strcasecmp (section->section_name, pp->pp_section))
		  continue;
		cfg_release_path (pp);
		return TRUE;
	}
	cfg_release_path (pp);
	return FALSE;
}

void * cfg_list_init(const gchar *path)
{
	cfg_profile   *New;
	cfg_section *section;
	PPath *pp;

	pp = cfg_parse_path (path);

	if (!cfg_is_loaded (pp->pp_file, &section)){
		New = (cfg_profile *) g_malloc (sizeof (cfg_profile));
		New->link = Base;
		New->filename = g_strdup (pp->pp_file);
		New->section = cfg_load (pp->pp_file);
		Base = New;
		section = New->section;
		Current = New;
	}
	for (; section; section = section->link){
		if (g_strcasecmp (section->section_name, pp->pp_section))
		  continue;
		cfg_release_path(pp);
		return section->keys;
	}
	cfg_release_path (pp);
	return 0;
}

void * cfg_list_next (void *s, gchar **key, gchar **value)
{
	cfg_keys *keys = (cfg_keys *) s;

	if (keys){
		*key   = g_strdup (keys->key_name);
		*value = g_strdup (keys->value);
		keys   = keys->link;
	}
	return keys;
}


/* EOF */
