#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <gtk/gtk.h>
#include "plop.h"
#include "pop3.h"

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif
#define DEFAULT_SOCKET_TIMEOUT 10
#define BUFFER_SIZE 128


/*
 * fonction d'ouverture d'un socket
 * renvoie -1 si erreur, sinon le socket descriptor
 */
int _socket(gchar *host, gchar *port)
{
	struct sockaddr_in addr_distante;
	struct hostent *h;
	gint desc;
	struct timeval tv;

	tv.tv_sec = DEFAULT_SOCKET_TIMEOUT;
	tv.tv_usec = 0;


#ifdef DEBUG
	fprintf(stderr, "--- Open socket... ");
#endif
   
	if ((desc = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
#ifdef DEBUG
		perror("socket");
#endif
		return -1;
	}
   
	//fcntl(desc, F_SETFL, fcntl(desc, F_GETFL) | O_NONBLOCK | O_ASYNC);
	/* timeout juste pour reception ? */
	setsockopt(desc, SOL_SOCKET, SO_RCVTIMEO, (void *)&tv, (socklen_t)sizeof(struct timeval));

	addr_distante.sin_family = AF_INET;
	addr_distante.sin_port = htons(atoi(port));

	if ((addr_distante.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE) {
		if (!(h = gethostbyname(host)))
			return -1;
		if (h->h_addr_list == NULL) {
#ifdef DEBUG
			perror("h_addr_list");
#endif
			return -1;
		}
	memcpy(&addr_distante.sin_addr, h->h_addr_list[0], h->h_length);
	}
   
	if (connect(desc, (struct sockaddr *) &addr_distante, sizeof addr_distante) < 0) {
#ifdef DEBUG
		perror("connect");
#endif
	      return -1;
	}
   
#ifdef DEBUG
	fprintf(stderr, "OK => connected at %s:%s\n", host, port);
#endif
   
	return desc;
}




/*
 * Fonction d'envoie d'une chaine
 * renvoie -1 si erreur, sinon 0
 */
int send_string(int socket, const char *string, ...)
{
	byte buffy[BUFFER_SIZE];
	va_list ap;

	bzero(buffy, sizeof buffy);
	va_start(ap, string);
	vsnprintf(buffy, (sizeof buffy)-1, string, ap);
	va_end(ap);

	if(send(socket, buffy, strlen(buffy), 0) == -1) {
#ifdef DEBUG
		printf("Erreur sending: %s\n", buffy);
		perror("send");
#endif
		return -1;
	}

#ifdef DEBUG
	printf(">>> %s", buffy);
	fflush(stdout);
#endif

	return 0;
	
}




/*
 * Fonction de reception d'une chaine
 * Verifie que le retour soit bien +OK
 * Sinon, verifie que c'est bien +ERR
 * Renvoie la chaine recu
 * NULL si erreur
 */
int recv_string(int socket, char *buf)
{
	gint len;
	byte buffy[BUFFER_SIZE];

	bzero(buffy, sizeof buffy);
	if((len =  recv(socket, buffy, (sizeof buffy)-1, 0)) == -1) {
#ifdef DEBUG
		perror("recv");
#endif
		return -1;
	}

	if(strncmp(buffy, "+OK", 3) != 0) {
		if(strncmp(buffy, "-ERR", 4) !=0) {
#ifdef DEBUG
			printf("UNKNOW ERROR RECV: %s\n", buffy);
			fflush(stdout);
#endif
			if(buf) bcopy(buffy, buf, len);
			return -2;
		}
#ifdef DEBUG
		printf("ERROR RECV: %s\n", buffy);
		fflush(stdout);
#endif
		if(buf) bcopy((const void*)buffy, (void *)buf, len);
		return -1;
	}
#ifdef DEBUG
	{
		int i = 0;
		while(buffy[i]) {
			if(!(i%70))
				printf("<<< ");
			if(buffy[i] > 31)
				printf("%c", buffy[i]);
			if(!((i+1)%70))
				printf("\n");
			i++;
		}
		puts(" ");
	}
	fflush(stdout);
#endif

	if(buf) bcopy(buffy, buf, len);
	return 0;
}


/*
 *
 */
int quit_mailbox(_POP_MAILBOX *mbox)
{
	if(send_string(mbox->socket, "QUIT\r\n"))
		return -1;

	if(recv_string(mbox->socket, NULL))
		return -1;

	return 0;
}

/*
 *
 */
int noop_mailbox(_POP_MAILBOX *mbox)
{
	if(send_string(mbox->socket, "NOOP\r\n"))
		return -1;

	if(recv_string(mbox->socket, NULL))
		return -1;

	return 0;
}

/*
 * Fonction d'identification au serveur
 * Envoie USER puis PASS. Essaie de quitter si erreur
 */
int user_mailbox(_POP_MAILBOX *mbox)
{
	if(!mbox->account->username)
		return -1;

	if(send_string(mbox->socket, "USER %s\r\n", mbox->account->username))
		return -1;
	if(recv_string(mbox->socket, NULL))
		return -1;

	return 0;
}

int pass_mailbox(_POP_MAILBOX *mbox)
{
	if(!mbox->account->password)
		return -1;

	if(send_string(mbox->socket, "PASS %s\r\n", mbox->account->password))
		return -1;
	if(recv_string(mbox->socket, NULL))
		return -1;

	return 0;
}





/*
 *
 */
int stat_mailbox(_POP_MAILBOX *mbox)
{
	char buffy[BUFFER_SIZE];
	bzero(buffy, sizeof buffy);

	if(send_string(mbox->socket, "STAT\r\n"))
		return -1;

	if(recv_string(mbox->socket, buffy))
		return -1;

	sscanf(buffy, "+OK %i %i", &mbox->number, &mbox->size);

	return 0;
}


/*
 *
 */
int list_mailbox(_POP_MAILBOX *mbox, gint *num, gint *size)
{
	char buffy[BUFFER_SIZE];
	bzero(buffy, sizeof buffy);

	if(*num) {
		if(send_string(mbox->socket, "LIST %i\r\n", *num))
			return -1;
	} else {
		if(send_string(mbox->socket, "LIST\r\n"))
			return -1;
	}

	if(recv_string(mbox->socket, buffy))
		return -1;

	if(*num) {
		*num = *size = 0;
		sscanf(buffy, "+OK %i %i\r\n", num, size);
	}
		
	return 0;
}

/*
 *
 */
int dele_mail(_POP_MAILBOX *mbox, gint num)
{
	if(send_string(mbox->socket, "DELE %i\r\n", num))
		return -1;
	if(recv_string(mbox->socket, NULL))
		return -1;

	return 0;
}


/*
 *
 */
int save_mail(gchar *b)
{

	return 0;
}

/*
 *
 */
char *top_mail(_POP_MAILBOX *mbox, gint num, gint size)
{
	char *buffy = (char *)malloc(size);
	bzero(buffy, size);

	if(send_string(mbox->socket, "TOP %i 0\r\n", num))
		return NULL;

	if(recv(mbox->socket, buffy, (size)-1, 0) == -1) {
#ifdef DEBUG
		perror("recv");
#endif
		return NULL;
	}

	if(strncmp(buffy, "+OK", 3) != 0) {
		if(strncmp(buffy, "-ERR", 4) !=0) {
#ifdef DEBUG
			printf("UNKNOW ERROR RECV: %s\n", buffy);
			fflush(stdout);
#endif
			return NULL;
		}
#ifdef DEBUG
		printf("ERROR RECV: %s\n", buffy);
		fflush(stdout);
#endif
		return NULL;
	}
#ifdef DEGROBUG
	{
		int i = 0;
		while(buffy[i]) {
			if((buffy[i] > 31) || (buffy[i] == '\n'))
				printf("%c", buffy[i]);
			i++;
		}
	}
	fflush(stdout);
#endif

	return buffy;
}


/*
 *
 */
char *retr_mail(_POP_MAILBOX *mbox, gint num, gint size)
{
	char *buffy = (char *)malloc(size);
	bzero(buffy, size);

	if(send_string(mbox->socket, "RETR %i\r\n", num))
		return NULL;

	if(recv(mbox->socket, buffy, (size)-1, 0) == -1) {
#ifdef DEBUG
		perror("recv");
#endif
		return NULL;
	}

	if(strncmp(buffy, "+OK", 3) != 0) {
		if(strncmp(buffy, "-ERR", 4) !=0) {
#ifdef DEBUG
			printf("UNKNOW ERROR RECV: %s\n", buffy);
			fflush(stdout);
#endif
			return NULL;
		}
#ifdef DEBUG
		printf("ERROR RECV: %s\n", buffy);
		fflush(stdout);
#endif
		return NULL;
	}
#ifdef DEGROBUG
	{
		int i = 0;
		while(buffy[i]) {
			if((buffy[i] > 31) || (buffy[i] == '\n'))
				printf("%c", buffy[i]);
			i++;
		}
	}
	fflush(stdout);
#endif

	return buffy;
}


/*
 *
 */
int retry_mailbox(_POP_MAILBOX *mbox)
{
	gchar *buffy = NULL;
	gint num = 1, size = 0;

	while(num <= mbox->number) {
		if(list_mailbox(mbox, &num, &size))
			return -1;
#ifdef DEBUG
		printf("--- LIST num: %i. size: %i\n", num, size);
#endif
		if(num) { // si il y a bien un mail
			if((buffy = top_mail(mbox, num, size))) {
				// check_header(buffy);
				if((buffy = retr_mail(mbox, num, size))) {
#ifdef DEBUG
					printf("--- RETR taille: %i\n", strlen(buffy));
#endif
					if(!save_mail(buffy)) { //si on peut pas sauver, on efface pas (sauf force ?)
						if(mbox->account->delete)
							dele_mail(mbox, num);
					} else {
						//error can't save
					}
				} else {
					// error can't retry
				}
			free(buffy);
			} else {
				//error can't retry header
			}
		}
		num++;
	} //while...

	return 0;
}

int check_mailbox(_POP_MAILBOX *mbox)
{
	/*
	_POP_ACCOUNT a = {
		"youplala",
		"POP3",
		"localhost",
		"110",
		"jym",
		"nghhy4df",
		0,
		1,
		10,
		NULL
	};
	_POP_MAILBOX mb = {
		&a,
		0,
		0,
		0,
		0,
		0,
	};

	_POP_MAILBOX *mbox = &mb;
	*/

	if(mbox->locked)
		return -1;

	mbox->locked = POP_MAILBOX_LOCKED;
	mbox->state = POP_MAILBOX_STATE_NOTHING;

	while(mbox->state != POP_MAILBOX_STATE_UPDATE) {
		switch(mbox->state) {
			case POP_MAILBOX_STATE_NOTHING:
#ifdef DEBUG
				puts("*** STATE_NOTHING");
#endif
				if(!mbox->socket)
					if((mbox->socket = _socket(mbox->account->host, mbox->account->port)) == -1)
						return -1;
				if(recv_string(mbox->socket, NULL)) {
					mbox->state = POP_MAILBOX_STATE_UPDATE;
					break;
				}
				mbox->state = POP_MAILBOX_STATE_AUTHORIZATION;
				break;
			case POP_MAILBOX_STATE_AUTHORIZATION:
#ifdef DEBUG
				puts("*** STATE_AUTH");
#endif
				if(user_mailbox(mbox)) {
					quit_mailbox(mbox);
					mbox->state = POP_MAILBOX_STATE_UPDATE;
					break;
				}
				if(pass_mailbox(mbox)) {
					quit_mailbox(mbox);
					mbox->state = POP_MAILBOX_STATE_UPDATE;
					break;
				}
				mbox->state = POP_MAILBOX_STATE_TRANSACTION;
				break;
			case POP_MAILBOX_STATE_TRANSACTION:
#ifdef DEBUG
				puts("*** STATE_TRANSAC");
#endif
				stat_mailbox(mbox);
#ifdef DEBUG
				printf("--- STAT num: %i, size: %i\n", mbox->number, mbox->size);
#endif
				if(mbox->number) {
					retry_mailbox(mbox);
				}

				quit_mailbox(mbox);
				mbox->state = POP_MAILBOX_STATE_UPDATE;
		}
	}
#ifdef DEBUG
	puts("*** STATE_UPDATE");
#endif
	close(mbox->socket);
	mbox->socket = 0;
	mbox->state = POP_MAILBOX_STATE_NOTHING;
	mbox->locked = POP_MAILBOX_FREE;
	return 0;

	return 0;
}
